"""Import an existing LAN Switch Interconnect into the subscription database."""

from uuid import uuid4

from orchestrator import step, workflow
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle
from orchestrator.workflow import StepList, begin, done
from orchestrator.workflows.steps import resync, set_status, store_process_subscription

from gso.cli.imports import LanSwitchInterconnectRouterSideImportModel, LanSwitchInterconnectSwitchSideImportModel
from gso.products import ProductName
from gso.products.product_blocks.lan_switch_interconnect import (
    LanSwitchInterconnectInterfaceBlockInactive,
    LanSwitchInterconnectRouterSideBlockInactive,
    LanSwitchInterconnectSwitchSideBlockInactive,
)
from gso.products.product_types.lan_switch_interconnect import ImportedLanSwitchInterconnectInactive
from gso.products.product_types.router import Router
from gso.products.product_types.switch import Switch
from gso.services.partners import get_partner_by_name
from gso.services.subscriptions import get_product_id_by_name
from gso.utils.types.virtual_identifiers import (
    DEFAULT_DCN_MANAGEMENT_VLAN_ID,
    DEFAULT_SWITCH_MANAGEMENT_VLAN_ID,
    VLAN_ID,
)


def _initial_input_form_generator() -> FormGenerator:
    class ImportLanSwitchInterconnect(FormPage):
        lan_switch_interconnect_description: str
        minimum_links: int
        switch_management_vlan_id: VLAN_ID | None = DEFAULT_SWITCH_MANAGEMENT_VLAN_ID
        dcn_management_vlan_id: VLAN_ID | None = DEFAULT_DCN_MANAGEMENT_VLAN_ID
        router_side: LanSwitchInterconnectRouterSideImportModel
        switch_side: LanSwitchInterconnectSwitchSideImportModel

    user_input = yield ImportLanSwitchInterconnect
    return user_input.model_dump() | {"partner": "GEANT"}


@step("Create subscription")
def create_subscription(partner: str) -> State:
    """Create a new subscription object."""
    partner_id = get_partner_by_name(partner)["partner_id"]
    product_id = get_product_id_by_name(ProductName.IMPORTED_LAN_SWITCH_INTERCONNECT)
    subscription = ImportedLanSwitchInterconnectInactive.from_product_id(product_id, partner_id)

    return {"subscription": subscription, "subscription_id": subscription.subscription_id}


@step("Initialize subscription")
def initialize_subscription(
    subscription: ImportedLanSwitchInterconnectInactive,
    lan_switch_interconnect_description: str,
    minimum_links: int,
    switch_management_vlan_id: VLAN_ID,
    dcn_management_vlan_id: VLAN_ID,
    router_side: dict,
    switch_side: dict,
) -> State:
    """Initialize the subscription using input data."""
    subscription.lan_switch_interconnect.lan_switch_interconnect_description = lan_switch_interconnect_description
    subscription.lan_switch_interconnect.minimum_links = minimum_links
    subscription.lan_switch_interconnect.switch_management_vlan_id = switch_management_vlan_id
    subscription.lan_switch_interconnect.dcn_management_vlan_id = dcn_management_vlan_id

    router_block = Router.from_subscription(router_side.pop("node")).router
    router_side_interfaces = [
        LanSwitchInterconnectInterfaceBlockInactive.new(uuid4(), **ae_member)
        for ae_member in router_side.pop("ae_members")
    ]
    subscription.lan_switch_interconnect.router_side = LanSwitchInterconnectRouterSideBlockInactive.new(
        uuid4(), **router_side, node=router_block, ae_members=router_side_interfaces
    )

    switch_block = Switch.from_subscription(switch_side.pop("switch")).switch
    switch_side_interfaces = [
        LanSwitchInterconnectInterfaceBlockInactive.new(uuid4(), **ae_member)
        for ae_member in switch_side.pop("ae_members")
    ]
    subscription.lan_switch_interconnect.switch_side = LanSwitchInterconnectSwitchSideBlockInactive.new(
        uuid4(), **switch_side, switch=switch_block, ae_members=switch_side_interfaces
    )

    return {"subscription": subscription}


@workflow(
    "Create Imported LAN Switch Interconnect", initial_input_form=_initial_input_form_generator, target=Target.CREATE
)
def create_imported_lan_switch_interconnect() -> StepList:
    """Create an imported LAN Switch Interconnect without provisioning it."""
    return (
        begin
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
