"""A creation workflow for creating a new interconnect between a switch and a router."""

from ipaddress import IPv4Network, IPv6Network
from typing import Annotated
from uuid import uuid4

from annotated_types import Len
from orchestrator.forms import FormPage
from orchestrator.targets import Target
from orchestrator.types import FormGenerator, State, SubscriptionLifecycle, UUIDstr
from orchestrator.utils.errors import ProcessFailureError
from orchestrator.workflow import StepList, begin, done, step, workflow
from orchestrator.workflows.steps import resync, set_status, store_process_subscription
from orchestrator.workflows.utils import wrap_create_initial_input_form
from pydantic import AfterValidator, ConfigDict
from pydantic_forms.validators import ReadOnlyField

from gso.products.product_blocks.lan_switch_interconnect import (
    LanSwitchInterconnectInterfaceBlockInactive,
)
from gso.products.product_types.lan_switch_interconnect import LanSwitchInterconnectInactive
from gso.products.product_types.router import Router
from gso.products.product_types.switch import Switch
from gso.services.infoblox import create_host_by_ip, create_v4_network_by_ip, create_v6_network_by_ip
from gso.services.partners import get_partner_by_name
from gso.settings import load_oss_params
from gso.utils.helpers import (
    active_router_selector,
    active_switch_selector,
    available_interfaces_choices,
    available_lags_choices,
    generate_lan_switch_interconnect_subnet_v4,
    generate_lan_switch_interconnect_subnet_v6,
)
from gso.utils.shared_enums import Vendor
from gso.utils.types.interfaces import (
    JuniperAEInterface,
    JuniperLAGMember,
    JuniperPhyInterface,
    LAGMember,
    PhysicalPortCapacity,
    validate_interface_names_are_unique,
)
from gso.utils.types.tt_number import TTNumber
from gso.utils.types.virtual_identifiers import DEFAULT_DCN_MANAGEMENT_VLAN_ID, DEFAULT_SWITCH_MANAGEMENT_VLAN_ID
from gso.workflows.shared import create_summary_form


def _initial_input_form(product_name: str) -> FormGenerator:
    class CreateLANSwitchInterconnectForm(FormPage):
        model_config = ConfigDict(title=product_name)

        tt_number: TTNumber
        partner: ReadOnlyField("GEANT", default_type=str)  # type: ignore[valid-type]
        router_side: active_router_selector()  # type: ignore[valid-type]
        switch_side: active_switch_selector()  # type: ignore[valid-type]
        description: str
        minimum_link_count: int

    initial_input = yield CreateLANSwitchInterconnectForm
    router = Router.from_subscription(initial_input.router_side)

    if router.router.vendor == Vendor.NOKIA:

        class NokiaLAGMemberA(LAGMember):
            interface_name: available_interfaces_choices(  # type: ignore[valid-type]
                router.subscription_id,
                PhysicalPortCapacity.TEN_GIGABIT_PER_SECOND,
            )

        router_side_ae_member_list = Annotated[
            list[NokiaLAGMemberA],
            AfterValidator(validate_interface_names_are_unique),
            Len(min_length=initial_input.minimum_link_count),
        ]
    else:
        router_side_ae_member_list = Annotated[  # type: ignore[assignment, misc]
            list[JuniperLAGMember],
            AfterValidator(validate_interface_names_are_unique),
            Len(min_length=initial_input.minimum_link_count),
        ]

    class InterconnectRouterSideForm(FormPage):
        model_config = ConfigDict(title="Please enter interface names and descriptions for the router side.")

        router_side_iface: available_lags_choices(initial_input.router_side) or JuniperAEInterface or str  # type: ignore[valid-type]
        router_side_ae_members: router_side_ae_member_list

    router_side_input = yield InterconnectRouterSideForm

    switch_side_ae_member_list = Annotated[
        list[JuniperLAGMember],
        AfterValidator(validate_interface_names_are_unique),
        Len(
            min_length=len(router_side_input.router_side_ae_members),
            max_length=len(router_side_input.router_side_ae_members),
        ),
    ]

    class InterconnectSwitchSideForm(FormPage):
        model_config = ConfigDict(title="Please enter interface names and descriptions for the switch side.")

        switch_side_iface: JuniperPhyInterface
        switch_side_ae_members: switch_side_ae_member_list

    switch_side_input = yield InterconnectSwitchSideForm
    user_input = initial_input.model_dump() | router_side_input.model_dump() | switch_side_input.model_dump()
    summary_fields = [
        "tt_number",
        "partner",
        "router_side",
        "switch_side",
        "description",
        "minimum_link_count",
        "router_side_iface",
        "router_side_ae_members",
        "switch_side_iface",
        "switch_side_ae_members",
    ]
    yield from create_summary_form(user_input, product_name, summary_fields)

    return user_input


@step("Create subscription")
def create_subscription(product: UUIDstr, partner: str) -> State:
    """Create a new subscription object in the database."""
    subscription = LanSwitchInterconnectInactive.from_product_id(product, get_partner_by_name(partner)["partner_id"])

    return {"subscription": subscription, "subscription_id": subscription.subscription_id}


@step("Initialize subscription")
def initialize_subscription(
    subscription: LanSwitchInterconnectInactive,
    description: str,
    minimum_link_count: int,
    router_side: UUIDstr,
    router_side_iface: JuniperPhyInterface,
    router_side_ae_members: list[dict],
    switch_side: UUIDstr,
    switch_side_iface: JuniperPhyInterface,
    switch_side_ae_members: list[dict],
) -> State:
    """Update the product model with all input from the operator."""
    subscription.lan_switch_interconnect.lan_switch_interconnect_description = description
    subscription.lan_switch_interconnect.minimum_links = minimum_link_count
    subscription.lan_switch_interconnect.switch_management_vlan_id = DEFAULT_SWITCH_MANAGEMENT_VLAN_ID
    subscription.lan_switch_interconnect.router_side.node = Router.from_subscription(router_side).router
    subscription.lan_switch_interconnect.router_side.ae_iface = router_side_iface
    for member in router_side_ae_members:
        subscription.lan_switch_interconnect.router_side.ae_members.append(
            LanSwitchInterconnectInterfaceBlockInactive.new(subscription_id=uuid4(), **member)
        )
    subscription.lan_switch_interconnect.switch_side.switch = Switch.from_subscription(switch_side).switch
    subscription.lan_switch_interconnect.switch_side.ae_iface = switch_side_iface
    for member in switch_side_ae_members:
        subscription.lan_switch_interconnect.switch_side.ae_members.append(
            LanSwitchInterconnectInterfaceBlockInactive.new(subscription_id=uuid4(), **member)
        )
    if subscription.lan_switch_interconnect.router_side.node.router_site.site_contains_optical_equipment:
        subscription.lan_switch_interconnect.dcn_management_vlan_id = DEFAULT_DCN_MANAGEMENT_VLAN_ID

    return {"subscription": subscription}


@step("Register IPv4 network in IPAM")
def register_dns_records_v4_network(subscription: LanSwitchInterconnectInactive) -> State:
    """Add DNS records in IPAM."""
    router_site = subscription.lan_switch_interconnect.router_side.node.router_site
    if not router_site or not router_site.site_internal_id:
        msg = "Site internal ID not set. Cannot continue."
        raise ProcessFailureError(msg, details=router_site)

    new_network = generate_lan_switch_interconnect_subnet_v4(router_site.site_internal_id)
    ipam_oss_params = load_oss_params().IPAM.LAN_SWITCH_INTERCONNECT
    create_v4_network_by_ip(
        ipam_oss_params.dns_view, ipam_oss_params.network_view, new_network, str(subscription.subscription_id)
    )

    return {"ipam_registrations": {"v4": {"network": new_network}}}


@step("Register IPv4 devices in IPAM")
def register_dns_records_v4_devices(
    subscription: LanSwitchInterconnectInactive, subscription_id: UUIDstr, ipam_registrations: dict[str, dict[str, str]]
) -> State:
    """Register DNS records for both switch and router side in IPAM."""
    switch_hostname = subscription.lan_switch_interconnect.switch_side.switch.fqdn
    router_hostname = (
        f"{subscription.lan_switch_interconnect.router_side.ae_iface}."
        f"{subscription.lan_switch_interconnect.router_side.node.router_fqdn}"
    )
    if not (switch_hostname and router_hostname):
        msg = "Missing switch or router hostname, cannot continue."
        raise ProcessFailureError(msg, details=subscription.lan_switch_interconnect)

    ip_network = IPv4Network(ipam_registrations["v4"]["network"])

    create_host_by_ip(switch_hostname, "LAN_SWITCH_INTERCONNECT", subscription_id, ipv4_address=ip_network[10])
    create_host_by_ip(router_hostname, "LAN_SWITCH_INTERCONNECT", subscription_id, ipv4_address=ip_network[1])

    return {"ipam_registrations": {"v4": {switch_hostname: ip_network[10], router_hostname: ip_network[1]}}}


@step("Register IPv6 network in IPAM")
def register_dns_records_v6_network(subscription: LanSwitchInterconnectInactive) -> State:
    """Add DNS records in IPAM."""
    router_site = subscription.lan_switch_interconnect.router_side.node.router_site
    if not router_site or not router_site.site_internal_id:
        msg = "Site internal ID not set. Cannot continue."
        raise ProcessFailureError(msg, details=router_site)

    new_network = generate_lan_switch_interconnect_subnet_v6(router_site.site_internal_id)
    ipam_oss_params = load_oss_params().IPAM.LAN_SWITCH_INTERCONNECT
    create_v6_network_by_ip(
        ipam_oss_params.dns_view, ipam_oss_params.network_view, new_network, str(subscription.subscription_id)
    )

    return {"ipam_registrations": {"v6": {"network": new_network}}}


@step("Register IPv6 devices in IPAM")
def register_dns_records_v6_devices(
    subscription: LanSwitchInterconnectInactive, subscription_id: UUIDstr, ipam_registrations: dict[str, dict[str, str]]
) -> State:
    """Register DNS records for both switch and router side in IPAM."""
    switch_hostname = subscription.lan_switch_interconnect.switch_side.switch.fqdn
    router_hostname = (
        f"{subscription.lan_switch_interconnect.router_side.ae_iface}."
        f"{subscription.lan_switch_interconnect.router_side.node.router_fqdn}"
    )
    if not (switch_hostname and router_hostname):
        msg = "Missing switch or router hostname, cannot continue."
        raise ProcessFailureError(msg, details=subscription.lan_switch_interconnect)

    ip_network = IPv6Network(ipam_registrations["v6"]["network"])

    create_host_by_ip(switch_hostname, "LAN_SWITCH_INTERCONNECT", subscription_id, ipv6_address=ip_network[10])
    create_host_by_ip(router_hostname, "LAN_SWITCH_INTERCONNECT", subscription_id, ipv6_address=ip_network[1])

    return {"ipam_registrations": {"v6": {switch_hostname: ip_network[10], router_hostname: ip_network[1]}}}


@workflow(
    "Create LAN Switch Interconnect",
    initial_input_form=wrap_create_initial_input_form(_initial_input_form),
    target=Target.CREATE,
)
def create_lan_switch_interconnect() -> StepList:
    """Create a new LAN interconnect between a Switch and a Router."""
    return (
        begin
        >> create_subscription
        >> store_process_subscription(Target.CREATE)
        >> initialize_subscription
        >> register_dns_records_v4_network
        >> register_dns_records_v4_devices
        >> register_dns_records_v6_network
        >> register_dns_records_v6_devices
        >> set_status(SubscriptionLifecycle.ACTIVE)
        >> resync
        >> done
    )
